//
//   mcfm_interface.cxx        
//
//                   
// 
//   Copyright (C) 2007 M.Sutton (sutt@cern.ch)    
//
//   $Id: mcfm_interfce.cxx, v   Fri  8 Nov 2013 09:07:01 CET sutt


#include <iostream>
#include <sstream>
#include <iomanip>
#include <string>
#include <sys/stat.h>

#include <cstdlib> 
#include <sys/time.h> 


#include "TFile.h"
#include "TH1D.h"
// #include "TMatrixT.h"
#include "TVectorT.h"
#include "TString.h"

#include "mcfm_grid.h"

  
// extern "C" struct {
//  bool creategrid;
//  int nSubProcess;
// } grid_;


// bool file_exists(const std::string& s) {   
//
//  if ( FILE* testfile=fopen(s.c_str(),"r") ) { 
//    fclose(testfile);
//    return true;
//  }
//  else return false;
// }




bool file_exists( const std::string& filename ) { 
  struct stat sb;
  if ( stat( filename.c_str(), &sb)==0 ) return true; // && S_ISREG(sb.st_mode ))
  else return false;
}


/// mcfm changed the the mxpart size between 6.7 and 6.8
// extern "C" int getmxpart_(void);

static const int mxpart = 14;    // mcfm parameter : max number of partons in event record. defined in Inc/constants.f

/// was 12 for mcfm 6.7
/// static const int mxpart = 12;    // mcfm parameter : max number of partons in event record. defined in Inc/constants.f

static const int _Ngrids = 5;
static       int  Ngrids = 5;
appl::mcfm_grid* mygrid[_Ngrids];

static const char* gridFiles[_Ngrids] = {
    "_eta3.root",
    "_eta4.root",
    "_pt34.root",
    "_rap34.root",
    "_m34.root"
};


static double Observable[_Ngrids] = { 0,  0,  0,  0,  0 };   // observable array
int             nObsBins[_Ngrids] = {11, 11, 26, 12, 10 }; // eta3, eta4, pt34, rap34, m34

/* For ATLAS7TeV Wetal 1612.03016*/
static const double _eta[12] = { 
  0.00, 0.21, 0.42, 0.63, 0.84, 
  1.05, 1.37, 1.52, 1.74, 1.95, 
  2.18, 2.50 };

/*For ATLAS7TeV ZpT 1406.3660*/
static const double _ptV[27] =  {
      0.0,   2.0,    4.0,    6.0,    8.0,
     10.0,  12.0,   14.0,   16.0,   18.0,
     22.0,  26.0,   30.0,   34.0,   38.0,
     42.0,  46.0,   50.0,   54.0,   60.0,
     70.0,  80.0,  100.0,  150.0,  200.0,
    300.0,  800.0
};
/*ATLAS8TeV high mass DY 1606.01736*/
static const double _rapV[13]={
0.0, 0.2, 0.4, 0.6, 0.8, 
1.0, 1.2, 1.4, 1.6, 1.8,
2.0, 2.2, 2.4
};
static const double _mV[11]={
66, 71, 76, 81, 86, 91, 96, 101, 106, 111, 116
};

  // number of observables and binning for observables  
  const double *obsBins[_Ngrids] = {_eta, _eta, _ptV, _rapV, _mV };

long unsigned int runs  =  0;
bool isBooked           =  false;
std::string glabel      =  "";

void getObservable( const double evt[][mxpart] );

int  cuts(int);


std::string date() { 
  time_t _t;
  time(&_t);
  return ctime(&_t);
}

void book_grid()  // inital grid booking
{
  if (isBooked) return;
  
  //  time_t _t;
  //  time(&_t);
  
  std::cout<<" ***********************************************"<<std::endl;
  std::cout<<" booking the grids " << date() << std::endl;
  
  // binning information for the grid constructor
  double xLow    = 1.0e-9, xUp = 1.0;
  int    nXbins  = 40;
  int    xorder  = 6;
  double q2Low   = 1.0*1.0;
  double q2Up    = 4000*4000;
  int    nQ2bins = 15;
  int    qorder  = 3;
  // set transform2 value
  double apramval=5.;
  appl::grid::transformvar(apramval);

  // lowest order in alphas	
  int lowest_order = 0;
  // how many loops
  int nloops = 1;
      
  std::string pdf_function;

  glabel = "mcfm_grid";

  const char* basename = std::getenv("appl_basename");
  if ( basename && std::string(basename)!="" ) glabel = basename;

  const char* q2upper = std::getenv("appl_q2up");
  if ( q2upper && std::string(q2upper)!="" ) q2Up = std::atof(q2upper);

  const char* q2lower = std::getenv("appl_q2low");
  if ( q2lower && std::string(q2lower)!="" ) q2Low = std::atof(q2lower);

  const char* q2order = std::getenv("appl_q2order");
  if ( q2order && std::string(q2order)!="" ) qorder = std::atoi(q2order);
 
//  std::cout << "q2low " << q2lower << "\tq2up " << q2upper << std::endl;
  std::cout << "q2low " << q2Low  << "\tq2up " << q2Up <<  std::endl;
  std::cout << "Process : " << nproc_.nproc << std::endl;

  if( nproc_.nproc == 1 )  
    {
      std::cout << " W+ production"; 
      pdf_function = "mcfmwp.config"; 
      glabel+="_Wp";

			double mw=80.385;
      q2Low   = pow(mw/2.,2), q2Up = pow(mw*2.,2);
      nQ2bins = 20;
      qorder  = 5;
			if(exptid_.exptid==10)
			{
				static const double Xbins[]={0.0, 1.0};
        nObsBins[0] = 1;
        obsBins[0] = Xbins;
        Ngrids  = 1;
        gridFiles[0]="_total_xsec.root";
			}
/*
      if(exptid_.exptid==11)
      {
         nXbins  = 10;
         xorder  = 6;
         glabel+="-10-6-3-1";
      } 
      if(exptid_.exptid==12)
      {
         nXbins  = 20;
         xorder  = 6;
         glabel+="-20-6-3-1";
      }  
      if(exptid_.exptid==13)
      {
         nXbins  = 30;
         xorder  = 6;
         glabel+="-30-6-3-1";
      }
      if(exptid_.exptid==14)
      {
         nXbins  = 40;
         xorder  = 6;
         glabel+="-40-6-3-1";
      }
*/
      if(exptid_.exptid==11)//setup for LHCb7Wp 1505.07024, LHCb8Wp 1511.08039
      {
          xLow = 1.4e-5;  xUp = 1.0;     nXbins = 30; xorder = 6;
         q2Low = 40*40;  q2Up = 160*160;nQ2bins = 10; qorder = 3;

         glabel+="_LHCb78";
         Ngrids  = 1;
         gridFiles[0]="_etal.root";
         static const double etal[]={2,2.25,2.5,2.75,3,3.25,3.5,4,4.5};
         nObsBins[0] = sizeof(etal)/sizeof(etal[0])-1;
         obsBins[0] = etal;
      }
      else if(exptid_.exptid==12)//setup for ATLAS7W 1612.03016
      {
          xLow = 3.0e-4;  xUp = 1.0;     nXbins = 30; xorder = 6;
         q2Low = 40*40;  q2Up = 160*160;nQ2bins = 10; qorder = 3;
    
         glabel+="_A7W";
         Ngrids  = 1;
         gridFiles[0]="_etal.root";
         static const double etal[]={0.00,0.21,0.42,0.63,0.84,1.05,1.37,1.52,1.74,1.95,2.18,2.50};
         nObsBins[0] = sizeof(etal)/sizeof(etal[0])-1;
         obsBins[0] = etal;
      }
      else if(exptid_.exptid==13)//setup for RHIC 510
      {
          xLow = 3.0e-4;  xUp = 1.0;     nXbins = 30; xorder = 6;
         q2Low = 40*40;  q2Up = 160*160;nQ2bins = 10; qorder = 3;
    
         glabel+="_RHIC510";
         Ngrids  = 1;
         gridFiles[0]="_etal.root";
         static const double etal[]={-1.00,-0.80,-0.50,-0.25,0.00,0.25,0.50,0.80,1.00,1.50};
         nObsBins[0] = sizeof(etal)/sizeof(etal[0])-1;
         obsBins[0] = etal;
      }
    }  
  else if ( nproc_.nproc == 6 )  
    {
      std::cout << " W- production"; 
      pdf_function = "mcfmwm.config"; 
      glabel+="_Wm";

      double mw=80.385;
      q2Low   = pow(mw/2.,2), q2Up = pow(mw*2.,2);
      nQ2bins = 20;
      qorder  = 5;

// ATLAS7W 1612.03016
      if(exptid_.exptid==60)
      {
        static const double Xbins[]={0.0, 1.0};
        nObsBins[0] = 1;
        obsBins[0] = Xbins;
        Ngrids  = 1;
        gridFiles[0]="_total_xsec.root";
      }
/*
      if(exptid_.exptid==61)
      {   
         nXbins  = 10; 
         xorder  = 6;
         glabel+="-10-6-3-1";
      }
      if(exptid_.exptid==62)
      {
         nXbins  = 20;
         xorder  = 6;
         glabel+="-20-6-3-1";
      }
      if(exptid_.exptid==63)
      {
         nXbins  = 30;
         xorder  = 6;
         glabel+="-30-6-3-1";
      }
      if(exptid_.exptid==64)
      {
         nXbins  = 40;
         xorder  = 6;
         glabel+="-40-6-3-1";
      }
*/
      else if(exptid_.exptid==61)//setup for LHCb7Wm 1505.07024, LHCb8Wm 1511.08039
      {
          xLow = 1.4e-5;  xUp = 1.0;     nXbins = 30; xorder = 6;
         q2Low = 40*40;  q2Up = 160*160;nQ2bins = 10; qorder = 3;

         glabel+="_LHCb78";
         Ngrids  = 1;
         gridFiles[0]="_etal.root";
         static const double etal[]={2,2.25,2.5,2.75,3,3.25,3.5,4,4.5};
         nObsBins[0] = sizeof(etal)/sizeof(etal[0])-1;
         obsBins[0] = etal;
      }
      if(exptid_.exptid==62)//setup for ATLAS7W 1612.03016
      {
          xLow = 3.0e-4;  xUp = 1.0;     nXbins = 30; xorder = 6;
         q2Low = 40*40;  q2Up = 160*160;nQ2bins = 10; qorder = 3;

         glabel+="_A7W";
         Ngrids  = 1;
         gridFiles[0]="_etal.root";
         static const double etal[]={0.00,0.21,0.42,0.63,0.84,1.05,1.37,1.52,1.74,1.95,2.18,2.50};
         nObsBins[0] = sizeof(etal)/sizeof(etal[0])-1;
         obsBins[0] = etal;
      }
      if(exptid_.exptid==63)//setup for RHIC510
      {
          xLow = 3.0e-4;  xUp = 1.0;     nXbins = 30; xorder = 6;
         q2Low = 40*40;  q2Up = 160*160;nQ2bins = 10; qorder = 3;

         glabel+="_RHIC510";
         Ngrids  = 1;
         gridFiles[0]="_etal.root";
         static const double etal[]={-1.00,-0.80,-0.50,-0.25,0.00,0.25,0.50,0.80,1.00,1.50};
         nObsBins[0] = sizeof(etal)/sizeof(etal[0])-1;
         obsBins[0] = etal;
      }

    }  
  else if      ( nproc_.nproc == 11 )  
    {
      std::cout << " W+ + jet production"; 
//        pdf_function = "mcfm-wpjet";
        pdf_function = "mcfm-zjet";
//	pdf_function = "basic";
        lowest_order = 1;
      glabel+="_WpJet";

      if(exptid_.exptid==111) //CMS8WZpT 1606.05864
      {
         xLow = 2.5e-4;   xUp = 1.0;        nXbins = 30; xorder  = 6;
         q2Low= 5.0*5.0; q2Up = 1200*1200; nQ2bins = 10; qorder  = 6;
         glabel+="_C8WZpT";
         Ngrids  = 1;
         gridFiles[0]="_ptwp.root";
         static const double ptz[] = {0,12.5,17.5,24,30,40,50,70,110,150,190,250,600};
         int num=sizeof(ptz)/sizeof(ptz[0])-1; nObsBins[0] = num;  obsBins[0] = ptz;
      }
    }
  else if ( nproc_.nproc == 16 )  
    {
      std::cout << " W- + jet production"; 
//      pdf_function = "mcfm-wmjet";
        pdf_function = "mcfm-zjet";
//        pdf_function = "basic";        
        lowest_order = 1; 
      glabel+="_WmJet";
      if(exptid_.exptid==161||exptid_.exptid==162) //CMS8WZpT 1606.05864
      {
         xLow = 2.5e-4;   xUp = 1.0;        nXbins = 30; xorder  = 6;
         q2Low= 5.0*5.0; q2Up = 1200*1200; nQ2bins = 10; qorder  = 6;
         glabel+="_C8WZpT";
         Ngrids  = 1;
         gridFiles[0]="_ptwm.root";
         static const double ptz[] = {0,12.5,17.5,24,30,40,50,70,110,150,190,250,600};
         int num=sizeof(ptz)/sizeof(ptz[0])-1; nObsBins[0] = num;  obsBins[0] = ptz;
      }
    }  
  else if ( nproc_.nproc == 31 ) 
    {
      std::cout << " Z production"; 
      pdf_function = "mcfm-z"; 
      glabel+="_Zonly";
      if(exptid_.exptid==310)
      {
				q2Low = pow(60./2., 2);
				q2Up  = pow(120.*2., 2);
        nQ2bins = 20;
        qorder  = 5;
				
        static const double Xbins[]={0.0, 1.0};
        nObsBins[0] = 1;
        obsBins[0] = Xbins;
        Ngrids  = 1;
        gridFiles[0]="_total_xsec.root";
      }
      else if(exptid_.exptid==311)//setup for CMS8DY 1412.1115
      {
         q2Low   = 7.0*7.0;
         q2Up    = 120*120;
         nQ2bins = 15;
         qorder  = 3;

         glabel+="_CMS8DY";
         Ngrids  = 2;
         gridFiles[0]="_Zlom.root";
         gridFiles[1]="_Zloy.root";
         static const double mzlo[]={15,20,25,30,35,40,45,50,55,60};
         static const double zrap[]={0.0,0.1,0.2,0.3,0.4,0.5,0.6,0.7,0.8,0.9,
                                     1.0,1.1,1.2,1.3,1.4,1.5,1.6,1.7,1.8,1.9,
                                     2.0,2.1,2.2,2.3,2.4};
         nObsBins[0] = sizeof(mzlo)/sizeof(mzlo[0])-1;
         nObsBins[1] = sizeof(zrap)/sizeof(zrap[0])-1;
         obsBins[0] = mzlo;
         obsBins[1] = zrap;
      }
      if(exptid_.exptid==312)//setup for CMS8DY 1412.1115
      { 
         q2Low   = 30.0*30.0;
         q2Up    = 240*240;
         nQ2bins = 15;
         qorder  = 3;

         glabel+="_CMS8DY";
         Ngrids  = 2;
         gridFiles[0]="_Zpkm.root";
         gridFiles[1]="_Zpky.root";
         static const double mzpk[]={60,64,68,72,76,81,86,91,96,101,106,110,115,120};
         static const double zrap[]={0.0,0.1,0.2,0.3,0.4,0.5,0.6,0.7,0.8,0.9,
                                     1.0,1.1,1.2,1.3,1.4,1.5,1.6,1.7,1.8,1.9,
                                     2.0,2.1,2.2,2.3,2.4};
         nObsBins[0] = sizeof(mzpk)/sizeof(mzpk[0])-1;
         nObsBins[1] = sizeof(zrap)/sizeof(zrap[0])-1;
         obsBins[0] = mzpk;
         obsBins[1] = zrap;
      }
      if(exptid_.exptid==313)//setup for CMS8DY 1412.1115
      { 
         q2Low   = 60.0*60.0;
         q2Up    = 400*400;
         nQ2bins = 15;
         qorder  = 3;

         glabel+="_CMS8DY";
         Ngrids  = 2;
         gridFiles[0]="_Zhim.root";
         gridFiles[1]="_Zhiy.root";
         static const double mzhi[]={120,126,133,141,150,160,171,185,200};
         static const double zrap[]={0.0,0.1,0.2,0.3,0.4,0.5,0.6,0.7,0.8,0.9,
                                     1.0,1.1,1.2,1.3,1.4,1.5,1.6,1.7,1.8,1.9,
                                     2.0,2.1,2.2,2.3,2.4};
         nObsBins[0] = sizeof(mzhi)/sizeof(mzhi[0])-1;
         nObsBins[1] = sizeof(zrap)/sizeof(zrap[0])-1;
         obsBins[0] = mzhi;
         obsBins[1] = zrap;
      }
      if(exptid_.exptid==314)//setup for CMS8DY 1412.1115
      {
         q2Low   = 100.0*100.0;
         q2Up    = 4000*4000;
         nQ2bins = 15;
         qorder  = 3;

         glabel+="_CMS8DY";
         Ngrids  = 2;
         gridFiles[0]="_Zxhim.root";
         gridFiles[1]="_Zxhiy.root";
         static const double mzhi[]={200,220,243,273,320,380,440,510,600,1000,1500,2000};
         static const double zrap[]={0.0,0.2,0.4,0.6,0.8,1.0,1.2,1.4,1.6,1.8,2.0,2.2,2.4};
         nObsBins[0] = sizeof(mzhi)/sizeof(mzhi[0])-1;
         nObsBins[1] = sizeof(zrap)/sizeof(zrap[0])-1;
         obsBins[0] = mzhi;
         obsBins[1] = zrap;
      }
      if(exptid_.exptid==315)//setup for LHCb7Zrap 1505.07024, LHCb8Zrap 1511.08039
      {
          xLow = 8e-5;  xUp = 1.0;     nXbins = 30; xorder = 6;
         q2Low = 30*30;q2Up = 240*240;nQ2bins = 10; qorder = 3;

         glabel+="_LHCb78";
         Ngrids  = 1;
         gridFiles[0]="_Zrap.root";
         static const double zrap[]={2,2.125,2.25,2.375,2.5,2.625,2.75,2.875,
                                     3,3.125,3.25,3.375,3.5,3.625,3.75,3.875,4,4.25,4.5};
         nObsBins[0] = sizeof(zrap)/sizeof(zrap[0])-1;
         obsBins[0] = zrap;
      }
      if(exptid_.exptid==316)//setup for A7Zrap 1612.03016, 46-66
      {
          xLow = 5.3e-4;  xUp = 1.0;    nXbins = 30; xorder = 6;
         q2Low = 23*23;q2Up = 132*132; nQ2bins = 10; qorder = 3;

         glabel+="_A7Zm46to66";
         Ngrids  = 1;
         gridFiles[0]="_Zrap.root";
         static const double zrap[]={0.0,0.4,0.8,1.2,1.6,2.0,2.4};
         nObsBins[0] = sizeof(zrap)/sizeof(zrap[0])-1;
         obsBins[0] = zrap;
      }
      if(exptid_.exptid==317)//setup for A7Zrap 1612.03016, 66-116
      {
          xLow = 7.7e-4;  xUp = 1.0;    nXbins = 30; xorder = 6;
         q2Low = 33*33;q2Up = 232*232; nQ2bins = 10; qorder = 3;

         glabel+="_A7Zm66to116";
         Ngrids  = 1;
         gridFiles[0]="_Zrap.root";
         static const double zrap[]={0.0,0.2,0.4,0.6,0.8,1.0,1.2,1.4,1.6,1.8,2.0,2.2,2.4};
         nObsBins[0] = sizeof(zrap)/sizeof(zrap[0])-1;
         obsBins[0] = zrap;
      }
      if(exptid_.exptid==318)//setup for A7Zrap 1612.03016, 66-116
      {
          xLow = 1.3e-3;  xUp = 1.0;    nXbins = 30; xorder = 6;
         q2Low = 58*58;q2Up = 300*300; nQ2bins = 10; qorder = 3;

         glabel+="_A7Zm116to150";
         Ngrids  = 1;
         gridFiles[0]="_Zrap.root";
         static const double zrap[]={0.0,0.4,0.8,1.2,1.6,2.0,2.4};
         nObsBins[0] = sizeof(zrap)/sizeof(zrap[0])-1;
         obsBins[0] = zrap;
      }
      if(exptid_.exptid==319)//setup for A7ZlowMDY 1404.1212
      {
          xLow = 1.5e-4; xUp = 1.0;    nXbins = 30; xorder = 6;
         q2Low = 6*6;   q2Up =132*132; nQ2bins = 10; qorder = 3;

         glabel+="_A7lowMDY";
         Ngrids  = 2;
         gridFiles[0]="_mll.root";
         static const double mll[]={26,31,36,41,46,51,56,61,66};
         nObsBins[0] = sizeof(mll)/sizeof(mll[0])-1;
         obsBins[0] = mll;
         gridFiles[1]="_mll1.root";
         static const double mll1[]={12,17,22,28,36,46,66};
         nObsBins[1] = sizeof(mll1)/sizeof(mll1[0])-1;
         obsBins[1] = mll1;
      }
      if(exptid_.exptid==3111)//setup for A8DY 1606.01736
      {
          xLow = 1.0e-3; xUp = 1.0;       nXbins = 30; xorder = 6;
         q2Low =58*58;  q2Up =3000*3000; nQ2bins = 10; qorder = 3;

         glabel+="_A8DY";
         Ngrids  = 3;
         gridFiles[0]="_mll.root";
         static const double mll[]={116,130,150,175,200,230,260,300,380,500,700,1000,1500};
         nObsBins[0] = sizeof(mll)/sizeof(mll[0])-1;
         obsBins[0] = mll;
         gridFiles[1]="_yll.root";
         static const double yll[]={0.0, 0.2, 0.4, 0.6, 0.8,1.0, 1.2, 1.4, 1.6, 1.8,2.0, 2.2, 2.4};
         nObsBins[1] = sizeof(yll)/sizeof(yll[0])-1;
         obsBins[1] = yll;
         gridFiles[2]="_yll1.root";
         static const double yll1[]={0.0, 0.4, 0.8, 1.2, 1.6, 2.0, 2.4};
         nObsBins[2] = sizeof(yll1)/sizeof(yll1[0])-1;
         obsBins[2] = yll1;
      }
      if(exptid_.exptid==3112)//setup for E866 hep-ex/0301031
      {
          xLow = 1.0e-3; xUp = 1.0; 
					nXbins = 30; xorder = 6;
					q2Low =4.2*4.2/4.0;  q2Up =16.85*16.85*4.0; 
					nQ2bins = 10; qorder = 5;

         glabel+="_E866";
         Ngrids  = 2;
         gridFiles[0]="_mll.root";
         static const double mll[]={4.2,4.7,5.2,5.7,6.2,6.7,7.2,7.7,8.2,8.7,10.85,11.85,12.85,14.85,16.85};
         nObsBins[0] = sizeof(mll)/sizeof(mll[0])-1;
         obsBins[0] = mll;
         gridFiles[1]="_xF.root";
         static const double xF[]={-0.05,0.05,0.10,0.15,0.20,0.25,0.30,0.35,0.40,
																				 0.45,0.50,0.55,0.60,0.65,0.70,0.75,0.80};
         nObsBins[1] = sizeof(xF)/sizeof(xF[0])-1;
         obsBins[1] = xF;
      }

    }  
  else if ( (nproc_.nproc >= 41) && (nproc_.nproc <= 43)) 
    {
      std::cout << " Z-jet production"; 
      pdf_function = "mcfm-zjet"; 
//	pdf_function = "basic";       
//      glabel+=TString::Format("-Zjet_%d",nproc_.nproc).Data();
			glabel+="_Zjet";      
			lowest_order = 1;
      if(exptid_.exptid>=411&&exptid_.exptid<=415)//CMS8ZpT 1504.03511
      {
         xLow  = 1.0e-3;    xUp = 1.0;       nXbins  = 30; xorder = 6; 
         q2Low = 40.0*40.0; q2Up= 1006*1006; nQ2bins = 10; qorder = 3;

         glabel+="_CMS8ZpT";
         Ngrids  = 1;
         gridFiles[0]="_ptz.root";
         static const double ptz[]  = {0,20,40,60,80,100,120,140,170,200,1000};
         int num=sizeof(ptz)/sizeof(ptz[0])-1;
         nObsBins[0] = num;
         obsBins[0] = ptz;
      }
      else if((exptid_.exptid==416)||(exptid_.exptid>=4161&&exptid_.exptid<=4166))//ATLAS8ZpT 1512.02192
      {
         xLow = 1.3e-4;   xUp = 1.0;        nXbins = 30; xorder  = 6;
         q2Low= 6.0*6.0; q2Up = 2000*2000; nQ2bins = 10; qorder  = 6;
         glabel+="_A8ZpT";
         Ngrids  = 2;
         gridFiles[0]="_ptz.root";  gridFiles[1]="_ptz1.root";  //gridFiles[2]="_ptz2.root";
         static const double ptz[]  = {0,30,33,36,39,42,45,48,51,54,57,61,65,70,75,80,85,95,
                                       105,125,150,175,200,250,300,350,400,470,550,650,900};
//         static const double ptz2[]  = {0,45,55,65,75,85,105,150,200,900};
         static const double ptz1[]  = {0,30,37,45,55,65,75,85,105,150,200,900};
         int num=sizeof(ptz)/sizeof(ptz[0])-1; nObsBins[0] = num;  obsBins[0] = ptz;
         int num1=sizeof(ptz1)/sizeof(ptz1[0])-1; nObsBins[1] = num1;  obsBins[1] = ptz1;
//         int num2=sizeof(ptz2)/sizeof(ptz2[0])-1; nObsBins[2] = num2;  obsBins[2] = ptz2;
      }
      else if((exptid_.exptid==417)||(exptid_.exptid>=4171&&exptid_.exptid<=4173))//ATLAS8ZpT 1406.3660
      {    
         xLow = 8.5e-4;    xUp = 1.0;        nXbins = 30; xorder  = 6; 
         q2Low=33.0*33.0; q2Up = 2000*2000; nQ2bins = 10; qorder  = 6; 
         glabel+="_A7ZpT";
         Ngrids  = 2; 
         gridFiles[0]="_ptz.root";  gridFiles[1]="_yz.root";
         static const double ptz[] = {0,18,22,26,30,34,38,42,46,50,54,60,70,80,100,150,200,300,800};
         static const double yz[]  = {0.0,1.0,2.0,2.4};
         int num=sizeof(ptz)/sizeof(ptz[0])-1; nObsBins[0] = num;  obsBins[0] = ptz;
         int num1=sizeof(yz)/sizeof(yz[0])-1;     nObsBins[1] = num1; obsBins[1] = yz;
      }
      else if(exptid_.exptid==418)//CMS8WZpT 1606.05864
      {
         xLow = 8.0e-4;   xUp = 1.0;        nXbins = 30; xorder  = 6;
         q2Low= 5.0*5.0; q2Up = 1200*1200; nQ2bins = 10; qorder  = 6;
         glabel+="_C8WZpT";
         Ngrids  = 1;
         gridFiles[0]="_ptz.root";
         static const double ptz[] = {0,10,12.5,15,17.5,20,30,40,50,70,90,110,150,190,250,600};
         int num=sizeof(ptz)/sizeof(ptz[0])-1; nObsBins[0] = num;  obsBins[0] = ptz;
      }

    }
//Single Top
  else if ( (nproc_.nproc >= 161) && (nproc_.nproc <= 177))
  {
         std::cout << " Single Top Production";
         pdf_function = "basic";
//         pdf_function = "mcfm-top";

         xLow   = 4.0e-10;
         xUp    = 1.0;
         nXbins  = 50;
         xorder  = 6;
         q2Low   = 80.0*80.0;
         q2Up    = 350.0*350.0;
         nQ2bins = 10;
         qorder  = 6;

         if(nproc_.nproc==161) glabel+="-SingleTop";
         if(nproc_.nproc==166) glabel+="-SingleTbr";
         if(nproc_.nproc==171) glabel+="-SingleTop-Schannel";
         if(nproc_.nproc==176) glabel+="-SingleTbr-Schannel";
         Ngrids  = 1;
         gridFiles[0]="_total_Xsect.root";
         static const double xsec[]  = {0,1};
         int num=sizeof(xsec)/sizeof(xsec[0])-1;
         nObsBins[0] = num;
         obsBins[0] = xsec; 
  } 
//W+top
  else if ((nproc_.nproc>=180)&&(nproc_.nproc<=187))
  {
         std::cout << " W+top production";
         pdf_function = "basic";
//         pdf_function = "mcfm-top";
//         lowest_order = 1;

         xLow   = 1.0e-7;
         xUp    = 1.0;
         nXbins  = 10;
         xorder  = 6;
         q2Low   = 170.0*170.0;
         q2Up    = 175.0*175.0;
         nQ2bins = 3;
         qorder  = 1;
         if(nproc_.nproc==180) glabel+="-WmTop";
         if(nproc_.nproc==185) glabel+="-WpTbr";

         Ngrids  = 1;
         gridFiles[0]="_total_Xsect.root";
         static const double xsec[]  = {0,1};
         int num=sizeof(xsec)/sizeof(xsec[0])-1;
         nObsBins[0] = num;
         obsBins[0] = xsec;
  }
  else if ( (nproc_.nproc >= 280) && (nproc_.nproc <= 286)) 
    {
      xLow = 2.8e-4;    xUp = 1.0;        nXbins = 30; xorder  = 6;
      q2Low= 25.0*25.0; q2Up = 1500*1500; nQ2bins= 10; qorder  = 6;

      Ngrids  = 5;
      gridFiles[0]="_etagamma.root";
      gridFiles[1]="_ptgfull.root";
      gridFiles[2]="_ptglow.root";
      gridFiles[3]="_ptgmid.root";
      gridFiles[4]="_ptghi.root";
      static const double eta[]={0.0,0.6,1.37,1.56,1.81,2.37};
      static const double ptg[]={25,35,45,55,65,75, 85,105,125,150,175,200,250,300,350,
                                 400,470,550,650,750, 900,1100,1500};
      static const double ptl[]={25,35,45,55,65,75, 85, 105};
      static const double ptm[]={105,125,150,175,200,250,300,350, 400};
      static const double pth[]={400,470,550,650,750, 900,1100,1500};
      nObsBins[0]=sizeof(eta)/sizeof(eta[0])-1; obsBins[0] = eta;
      nObsBins[1]=sizeof(ptg)/sizeof(ptg[0])-1; obsBins[1] = ptg;
      nObsBins[2]=sizeof(ptl)/sizeof(ptl[0])-1; obsBins[2] = ptl;
      nObsBins[3]=sizeof(ptm)/sizeof(ptm[0])-1; obsBins[3] = ptm;
      nObsBins[4]=sizeof(pth)/sizeof(pth[0])-1; obsBins[4] = pth;

      std::cout << " Photon production"; 
//      pdf_function = "photonLO.config:photonNLO.config";
//      pdf_function = "mcfm-zjet";
      pdf_function = "basic";
      glabel += TString::Format("-GammaProd_%d",nproc_.nproc).Data();
    }  
  else if ( nproc_.nproc == 13 )  
    {
      std::cout << " W+ + Cbar production"; 
      pdf_function = "mcfm-wpc"; 
      glabel+="-WpCbar";
      lowest_order = 1;
      if(exptid_.exptid==131)//ATLAS7Wcharm 1402.6263
      {
         xLow  = 5.0e-4;    xUp = 1.0;     nXbins  = 30; xorder = 6;
         q2Low = 40.0*40.0; q2Up= 160*160; nQ2bins =  5; qorder = 3;
         glabel+="-ATLAS7";
         Ngrids  = 1;
         gridFiles[0]="_etal.root";
         static const double eta[]  = {0,0.21,0.42,0.63,0.84,1.05,1.37,1.52,1.74,1.95,2.18,2.50};
         int num=sizeof(eta)/sizeof(eta[0])-1;
         nObsBins[0] = num;
         obsBins[0] = eta; 
       }
      if(exptid_.exptid==132)//CMS7Wcharm 1310.1138
      {
         xLow  = 5.0e-4;    xUp = 1.0;     nXbins  = 30; xorder = 6;
         q2Low = 40.0*40.0; q2Up= 160*160; nQ2bins =  5; qorder = 3;
         glabel+="-CMS7";
         Ngrids  = 1;
         gridFiles[0]="_etal.root";
         static const double eta[]  = {0,0.35,0.70,1.10,1.60,2.10};
         int num=sizeof(eta)/sizeof(eta[0])-1;
         nObsBins[0] = num;
         obsBins[0] = eta;
       }
    }  
  else if ( nproc_.nproc == 18 )  
    {
      std::cout << " W- + C production"; 
      pdf_function = "mcfm-wmc"; 
      glabel+="-WmCjet";      
      lowest_order = 1;
      if(exptid_.exptid==181)//ATLAS7Wcharm 1402.6263
      {
         xLow  = 5.0e-4;    xUp = 1.0;     nXbins  = 30; xorder = 6;
         q2Low = 40.0*40.0; q2Up= 160*160; nQ2bins =  5; qorder = 3;
         glabel+="-ATLAS7";
         Ngrids  = 1;
         gridFiles[0]="_etal.root";
         static const double eta[]  = {0,0.21,0.42,0.63,0.84,1.05,1.37,1.52,1.74,1.95,2.18,2.50};
         int num=sizeof(eta)/sizeof(eta[0])-1;
         nObsBins[0] = num;
         obsBins[0] = eta;
      }
      if(exptid_.exptid==182)//CMS7Wcharm 1310.1138
      {    
         xLow  = 5.0e-4;    xUp = 1.0;     nXbins  = 30; xorder = 6; 
         q2Low = 40.0*40.0; q2Up= 160*160; nQ2bins =  5; qorder = 3; 
         glabel+="-CMS7";
         Ngrids  = 1; 
         gridFiles[0]="_etal.root";
         static const double eta[]  = {0,0.35,0.70,1.10,1.60,2.10};
         int num=sizeof(eta)/sizeof(eta[0])-1;
         nObsBins[0] = num; 
         obsBins[0] = eta; 
       } 
    }
  else if ( (nproc_.nproc == 141) || (nproc_.nproc == 142) || (nproc_.nproc==144) ||
	    (nproc_.nproc == 145) || (nproc_.nproc == 146) || (nproc_.nproc==147) ||
	    (nproc_.nproc == 148) || (nproc_.nproc == 149) || (nproc_.nproc==150) ||(nproc_.nproc == 151) ||
	    (nproc_.nproc == 157) || (nproc_.nproc == 158) || (nproc_.nproc==159) ||
            (nproc_.nproc >= 701 && nproc_.nproc <= 704) )
    {
      lowest_order = 2;
      
      if (nproc_.nproc == 141)
	{
	  std::cout << " TTbar production with 2 semi-leptonic decays"; 
	  pdf_function = "mcfm-TT"; 
	  glabel+="-TTbar-141";
	}
      else if (nproc_.nproc == 142)
	{
	  std::cout << " TTbar production with 2 semi-leptonic decays, corrections only in decays"; 
	  pdf_function = "mcfm-TT"; 
	  glabel+="-TTbar-142";
	}
      else if (nproc_.nproc == 144)
	{
	  std::cout << " TTbar production with 2 semi-leptonic decays, no correlations"; 
	  pdf_function = "mcfm-TT"; 
	  glabel+="-TTbar-144";
	}
      else if (nproc_.nproc == 145)
	{
	  std::cout << " TTbar production with 2 semi-leptonic decays, no spin correlations in top decays"; 
	  pdf_function = "mcfm-TT"; 
	  glabel+="-TTbar-145";
	}
      else if (nproc_.nproc == 146)
	{
	  std::cout << " TTbar production with Tbar hadronic decay, radiative corrections in production and decay"; 
	  pdf_function = "mcfm-TT"; 
	  glabel+="-TTbar-146";
	}
      else if (nproc_.nproc == 147)
	{
	  std::cout << " TTbar production with Tbar hadronic decay, radiative corrections in Tbar decay"; 
	  pdf_function = "mcfm-TT"; 
	  glabel+="-TTbar-147";
	}
      else if (nproc_.nproc == 148)
	{
	  std::cout << " TTbar production with Tbar hadronic decay, radiative corrections in W decay"; 
	  pdf_function = "mcfm-TT"; 
	  glabel+="-TTbar-148";
	}
      else if (nproc_.nproc == 149)
	{
	  std::cout << " TTbar production with T hadronic decay, radiative corrections in production and decay"; 
	  pdf_function = "mcfm-TT"; 
	  glabel+="-TTbar-149";
	}
      else if (nproc_.nproc == 150)
	{
	  std::cout << " TTbar production with T hadronic decay, radiative corrections in T decay"; 
	  pdf_function = "mcfm-TT"; 
	  glabel+="-TTbar-150";
	}
      else if (nproc_.nproc == 151)
	{
	  std::cout << " TTbar production with T hadronic decay, radiative correstions in W decay"; 
	  pdf_function = "mcfm-TT"; 
	  glabel+="-TTbar-151";
	}
      else if (nproc_.nproc == 157)
	{
	  std::cout << " TTbar production"; 
	  pdf_function = "mcfm-TT"; 
	  glabel+="-TTbar";
	}
      else if(nproc_.nproc == 158||(nproc_.nproc >= 701 && nproc_.nproc <= 704))
	{
          if(nproc_.nproc == 158){
     	   std::cout << " BBbar production"; 
 	   pdf_function = "mcfm-BB"; 
	   glabel+="-BBbar";  
          }
          else if(nproc_.nproc >= 701 && nproc_.nproc <= 704){
//           std::cout << "bf->bf production";
           pdf_function = "mcfm-Bl";
//             pdf_function = "mcfm-bgq";
//           pdf_function = "basic";
           glabel+="-BX";
             
          }
          lowest_order = 2;
          if(exptid_.exptid==1581||(nproc_.nproc >= 701 && nproc_.nproc <= 704))
          {
             xLow  = 1.0e-6;    xUp = 1.0;     nXbins  = 50; xorder = 6;
             q2Low = 2*2; q2Up= 85*85; nQ2bins =  30; qorder = 6;
             glabel+="-L7v13";
             Ngrids  = 2;
             gridFiles[0]="_pTb.root";
             gridFiles[1]="_yb.root";
//             gridFiles[2]="_pTby1.root"; gridFiles[3]="_pTby2.root"; gridFiles[4]="_pTby3.root";
//             gridFiles[5]="_pTby4.root"; gridFiles[6]="_pTby5.root";
             static const double pTb[] ={0.0,0.5,1.0,1.5,2.0,2.5,3.0,3.5,4.0,4.5,5.0,5.5,6.0,6.5,
        7.0,7.5,8.0,8.5,9.0,9.5,10.0,10.5,11.5,12.5,14.0,16.5,23.5,40.0};
             static const double yb[]  = {2.0,2.5,3.0,3.5,4.0,4.5};
             int num1=sizeof(pTb)/sizeof(pTb[0])-1;
             int num2=sizeof( yb)/sizeof( yb[0])-1;
             nObsBins[0] = num1; obsBins[0] = pTb;
             nObsBins[1] = num2; obsBins[1] =  yb;
/*             nObsBins[2] = num1; obsBins[2] = pTb;
             nObsBins[3] = num1; obsBins[3] = pTb;
             nObsBins[4] = num1; obsBins[4] = pTb;
             nObsBins[5] = num1; obsBins[5] = pTb;
             nObsBins[6] = num1; obsBins[6] = pTb;
*/
           }
	}
      else if (nproc_.nproc == 159)
	{
	  std::cout << " CCbar production"; 
	  pdf_function = "mcfm-CC"; 
	  glabel+="-CCbar";
	}
    }  
  else                           
    { 
      std::cerr << "don't know which process" << std::endl; 
      std::exit(-1); 
    } 
  std::cout << std::endl;

  /// Read the ckm matrix from mcfm to store in the grid automatically
  /// NB: we store 13 x 13 ckm matrix - mcfm only stores 11 x 11 so we 
  ///     must add 1 to each index to keep them aligned

  std::vector< std::vector<double> > ckm_vsq( 13, std::vector<double>( 13, 0 ) );
  
  for ( int ic=0 ; ic<__nf2__ ; ic++ ) { 
    for ( int ic1=0 ; ic1<__nf2__ ; ic1++ ) ckm_vsq[ic+1][ic1+1] = ckm_.vsq[ic][ic1];
  } 

  std::vector<std::vector<double> > __ckm( 3, std::vector<double>(3, 0) );
  __ckm[0][0] = cabib_.Vud;
  __ckm[0][1] = cabib_.Vus;
  __ckm[0][2] = cabib_.Vub;
  __ckm[1][0] = cabib_.Vcd;
  __ckm[1][1] = cabib_.Vcs;
  __ckm[1][2] = cabib_.Vcb;


  for(int igrid=0; igrid < Ngrids; igrid++) 
    {
    
      bool create_new = false;

      // if the file does not exist, create a new grid...
      if ( !file_exists(glabel+gridFiles[igrid]) )  create_new = true;

      // or if it does exists but root file is a zombie...
      if ( !create_new ) {  
	TFile testFile( (glabel+gridFiles[igrid]).c_str() );
	if ( testFile.IsZombie() ) create_new = true;
	testFile.Close();
      }

      if ( create_new ) 
	{ 
	  std::cout << "Creating NEW grid... " << std::endl;
	  
	  std::cout << "grid interpolation: " 
		    << "\tQ2 " << nQ2bins << " " <<  q2Low << " " <<  q2Up << " " <<  qorder   
		    << "\tx "  <<  nXbins << " " <<   xLow << " " <<   xUp << " " <<  xorder
		    << std::endl; 
	    


	  mygrid[igrid] = new appl::mcfm_grid( nObsBins[igrid], obsBins[igrid],      // obs bins
					       nQ2bins, q2Low, q2Up, qorder,         // Q2 bins and interpolation order
					       nXbins,   xLow,  xUp, xorder,         // x bins and interpolation order
					       pdf_function, lowest_order, nloops ); 
	  /// try reweighting for a bit
	  mygrid[igrid]->reweight(true);
// turn the reweight off
//	mygrid[igrid]->reweight(false);	  
	  mygrid[igrid]->setCMSScale( energy_.sqrts );


	  /// store the ckm matrix
	  //	  mygrid[igrid]->setckm2( ckm_vsq );

	  mygrid[igrid]->setckm( __ckm );

	  //	  grid_.nSubProcess = mygrid[igrid]->subProcesses();
	  
	  std::cout << "reference histo name = " 
	       << mygrid[igrid]->getReference()->GetName() << std::endl;
	  
//	  std::cout<<*mygrid[igrid]<<std::endl;  
	}
      else 
	{
	  std::cout << "Using existing grid file " << (glabel+gridFiles[igrid]) << std::endl;
	  
	  mygrid[igrid] = new appl::mcfm_grid(glabel+gridFiles[igrid]); //optimise grid x,Q2 bins
	  //       grid_.nSubProcess = mygrid[igrid]->subProcesses();
	  mygrid[igrid]->getReference()->Reset();
	  mygrid[igrid]->optimise(nQ2bins, nXbins);
	  
//	  std::cout<<*(mygrid[igrid])<<std::endl;  
	}
      // CTEQ like reweighting
      //      mygrid[igrid]->reweight( false );
    }

  runs = 0;
  isBooked = true;
  std::cout<<" ***********************************************"<<std::endl;
}


void fill_grid( const double evt[][mxpart] )
{

  static unsigned evtcounter = 1;
  if ( evtcounter%50000==0 ) std::cout << "fill_grid() filled " << evtcounter << " weights " << date(); 
  evtcounter++;

  if (!isBooked) 
    {    
      book_grid();
      return;
    }

  getObservable( evt );
  
  for(int igrid = 0; igrid < Ngrids; igrid++)
    if(cuts(igrid)){
 
      mygrid[igrid]->fillMCFM( Observable[igrid] );
    }

  runs++; // counter of number of events (shouldn't this be after cuts)? or is it the number of runs?"
}


//
// just normalise to bin width
//
void Normalise(TH1D* h) 
{ 
  for ( int ibin=1 ; ibin<=h->GetNbinsX() ; ibin++ ) 
    { 
      double width = h->GetBinLowEdge(ibin+1) - h->GetBinLowEdge(ibin);
      h->SetBinContent( ibin, h->GetBinContent(ibin)/width );
    }
  return;
}



void write_grid(double& xstotal)   // writes out grid after some events
{
  std::cout<<"Write out grids ..."<<std::endl;

  for(int igrid = 0; igrid < Ngrids; igrid++)
    {
      std::cout << "saving grid N=" << igrid+1 << "\tof " << Ngrids << "\t";

      std::system("sleep 1");
      

      mygrid[igrid]->setNormalised( false );
      mygrid[igrid]->run() = (iterat_.ncall2)*(iterat_.itmx2);
      
      mygrid[igrid]->untrim();
      int untrim_size = mygrid[igrid]->size();

      mygrid[igrid]->trim();
      int trim_size = mygrid[igrid]->size();

      /// scale up by number of weights
      (*mygrid[igrid]) *= mygrid[igrid]->run();
      
      // normalise the reference histogram by bin width
      Normalise( mygrid[igrid]->getReference() );

      /// now scale *down* the reference histogram because we've just 
      /// scaled it up ...
      //      mygrid[igrid]->getReference()->Scale( 1/mygrid[igrid]->run() );

      std::string filename = glabel+gridFiles[igrid];

#if 0

      std::string newpdfname = "";

      /// automatically optimise subprocesses - doesn't quite work yet
      /// when it does it may be moved into the grid itself
      if ( mygrid[igrid]->getGenpdf().find("basic")!=std::string::npos ) { 
	
       	std::stringstream ss;
	ss << "proc" << nproc_.nproc;
	newpdfname = ss.str();

	std::cout << "appl::grid::Write() " << newpdfname << std::endl;

	mygrid[igrid]->Write( filename, "grid", newpdfname );
      }
      else { 
	mygrid[igrid]->Write( glabel+gridFiles[igrid] );
      }

#else

      mygrid[igrid]->Write( filename );

#endif


      std::cout << "size(untrimmed)=" << untrim_size 
		<< "\tsize(trimmed)=" << trim_size 
		<< "\tfraction="      << 100.*trim_size/untrim_size << " %" << std::endl;

      //      int nsub = mygrid[igrid]->subProcesses();

      delete mygrid[igrid];
      
    }
  
  time_t _t;
  time(&_t);
  
  std::cout<<" ***********************************************"<<std::endl;
  std::cout<<" saved grids " << ctime(&_t);
  std::cout<<" ***********************************************"<<std::endl;
}
 


//
// ----------------------------------------------
//    analysis
// ----------------------------------------------
//

void getObservable(const double evt[][mxpart])
{
  // evt[momentum][particle number-1]
  // momentum[0,1,2,3] = (x,y,z,E)
  //

  // calculate observables
  for(int igrid = 0; igrid < Ngrids; igrid++)Observable[igrid] = 0.0; // initialize
  double p1[4] = {evt[3][0],evt[0][0],evt[1][0],evt[2][0]};
	double p2[4] = {evt[3][1],evt[0][1],evt[1][1],evt[2][1]};
  double p3[4] = {evt[3][2],evt[0][2],evt[1][2],evt[2][2]}; // (E,x,y,z)
  double p4[4] = {evt[3][3],evt[0][3],evt[1][3],evt[2][3]};
  double p5[4] = {evt[3][4],evt[0][4],evt[1][4],evt[2][4]};
  double p6[4] = {evt[3][5],evt[0][5],evt[1][5],evt[2][5]};
  double p7[4] = {evt[3][6],evt[0][6],evt[1][6],evt[2][6]};
  double p8[4] = {evt[3][7],evt[0][7],evt[1][7],evt[2][7]};

  double rapidity3 = 0.0;
  rapidity3 = (p3[0] + p3[3])/(p3[0] - p3[3]);
  (rapidity3 < 1e-13) ? rapidity3 = 100.0 : rapidity3 = 0.5*std::log(rapidity3);
  
  double rapidity4 = 0.0;
  rapidity4 = (p4[0] + p4[3])/(p4[0] - p4[3]);
  (rapidity4 < 1e-13) ? rapidity4 = 100.0 : rapidity4 = 0.5*std::log(rapidity4);

  double rapidity5 = 0.0;
  rapidity5 = (p5[0] + p5[3])/(p5[0] - p5[3]);
  (rapidity5 < 1e-13) ? rapidity3 = 100.0 : rapidity5 = 0.5*std::log(rapidity5);

  double rapidity6 = 0.0;
  rapidity6 = (p6[0] + p6[3])/(p6[0] - p6[3]);
  (rapidity6 < 1e-13) ? rapidity6 = 100.0 : rapidity6 = 0.5*std::log(rapidity6);

  double rapidity7 = 0.0;
  rapidity7 = (p7[0] + p7[3])/(p7[0] - p7[3]);
  (rapidity7 < 1e-13) ? rapidity7 = 100.0 : rapidity7 = 0.5*std::log(rapidity7);
  
  double rapidity8 = 0.0;
  rapidity8 = (p8[0] + p8[3])/(p8[0] - p8[3]);
  (rapidity8 < 1e-13) ? rapidity8 = 100.0 : rapidity8 = 0.5*std::log(rapidity8);
  
  double rapidity34 = 0.0;                      // rapidity of particle (3+4) in event record
  rapidity34  = (p3[0] + p4[0]) + (p3[3] + p4[3]);
  rapidity34 /= (p3[0] + p4[0]) - (p3[3] + p4[3]);  
  (rapidity34 < 1e-13) ? rapidity34 = 100.0 : rapidity34 = 0.5*std::log(rapidity34);

  double rapidity56 = 0.0;                      // rapidity of particle (5+6) in event record
  rapidity56  = (p5[0] + p6[0]) + (p5[3] + p6[3]);
  rapidity56 /= (p5[0] + p6[0]) - (p5[3] + p6[3]);
  (rapidity56 < 1e-13) ? rapidity56 = 100.0 : rapidity56 = 0.5*std::log(rapidity56);

  double rapidity345 = 0.0;                      // rapidity of particle (3+4+5) in event record
  rapidity345  = (p3[0] + p4[0] + p5[0]) + (p3[3] + p4[3] + p5[3]);
  rapidity345 /= (p3[0] + p4[0] + p5[0]) - (p3[3] + p4[3] + p5[3]);  
  (rapidity345 < 1e-13) ? rapidity345 = 100.0 : rapidity345 = 0.5*std::log(rapidity345);

  double rapidity678 = 0.0;                      // rapidity of particle (7+6+8) in event record
  rapidity678  = (p6[0] + p7[0] + p8[0]) + (p6[3] + p7[3] + p8[3]);
  rapidity678 /= (p6[0] + p7[0] + p8[0]) - (p6[3] + p7[3] + p8[3]);  
  (rapidity678 < 1e-13) ? rapidity678 = 100.0 : rapidity678 = 0.5*std::log(rapidity678);

  
  double pt3 = 0;
  pt3 = std::sqrt( p3[1]*p3[1] + p3[2]*p3[2] );
  
  double pt4 = 0;
  pt4 = std::sqrt( p4[1]*p4[1] + p4[2]*p4[2] );

  double pt5 = 0;
  pt5 = std::sqrt( p5[1]*p5[1] + p5[2]*p5[2] );

  double pt6 = 0;
  pt6 = std::sqrt( p6[1]*p6[1] + p6[2]*p6[2] );
  
  double pt7 = 0;
  pt7 = std::sqrt( p7[1]*p7[1] + p7[2]*p7[2] );

  double pt8 = 0;
  pt8 = std::sqrt( p8[1]*p8[1] + p8[2]*p8[2] );
  
  double pt34 = 0;
  pt34 = std::sqrt( std::pow(p3[1] + p4[1],2) + std::pow(p3[2] + p4[2],2) );

  double pt56 = 0;
  pt56 = std::sqrt( std::pow(p5[1] + p6[1],2) + std::pow(p5[2] + p6[2],2) );
  
  double pt345 = 0;
  pt345 = std::sqrt( std::pow(p3[1] + p4[1] + p5[1],2) + std::pow(p3[2] + p4[2] + p5[2],2) );

  double pt678 = 0;
  pt678 = std::sqrt( std::pow(p6[1] + p7[1] + p8[1],2) + std::pow(p6[2] + p7[2] + p8[2],2) );

  double m34 = 0;
  m34 = m34 + std::pow(p3[0] + p4[0],2);
  m34 = m34 - std::pow(p3[1] + p4[1],2);
  m34 = m34 - std::pow(p3[2] + p4[2],2);
  m34 = m34 - std::pow(p3[3] + p4[3],2);
  m34 = std::sqrt(m34); 
	double sqrts=std::pow(p1[0] + p2[0],2)-std::pow(p1[1] + p2[1],2);
	sqrts =sqrts-std::pow(p1[2] + p2[2],2)-std::pow(p1[3] + p2[3],2);
	sqrts =std::sqrt(sqrts);
	double pzll=p3[3] + p4[3];
	double xL=2*pzll/sqrts;

	double Xval=0.5;
	if(exptid_.exptid%10==0) {
		Observable[ 0 ] = Xval;
	}
  else if ( (nproc_.nproc >= 280) && (nproc_.nproc <= 286)) {
    Observable[ 0 ] = std::fabs(rapidity3);
    Observable[ 1 ] = pt3;
    Observable[ 2 ] = pt3;
    Observable[ 3 ] = pt3;
    Observable[ 4 ] = pt3;
  }
  // nproc=1 W+; nproc=6 W-
  else if((nproc_.nproc==1)){
   if(exptid_.exptid==101) Observable[ 0 ] = rapidity4;
   else Observable[ 0 ] = std::fabs(rapidity4);
  }
  else if((nproc_.nproc==6)){
   if(exptid_.exptid==601) Observable[ 0 ] = rapidity3;
   else Observable[ 0 ] = std::fabs(rapidity3);
  }
//W+jet
  else if((nproc_.nproc==11)||(nproc_.nproc==16)){
   Observable[ 0 ] = pt34;
  }
  //Z+jet
  else if( (exptid_.exptid>=411&&exptid_.exptid<=415)||(exptid_.exptid==418)){
    Observable[ 0 ] = pt34;
  }
  else if( (exptid_.exptid==416)||(exptid_.exptid>=4161&&exptid_.exptid<=4166)){
    Observable[ 0 ] = pt34;
    Observable[ 1 ] = pt34;
//    Observable[ 2 ] = pt34;
  }
  else if( (exptid_.exptid==417)||(exptid_.exptid>=4171&&exptid_.exptid<=4173)){
    Observable[ 0 ] = pt34;
    Observable[ 1 ] = std::fabs(rapidity34);
  }
  //Z
  else if((exptid_.exptid>=311)&&(exptid_.exptid<=314)){
    Observable[ 0 ] = m34;
    Observable[ 1 ] = std::fabs(rapidity34);
  }
//LHCb78
  else if(exptid_.exptid>=315&&exptid_.exptid<=318){
    Observable[ 0 ] = std::fabs(rapidity34);
  }
//A7lowMDY
  else if(exptid_.exptid==319){
    Observable[ 0 ] = m34;
    Observable[ 1 ] = m34;
  }
  else if(exptid_.exptid==3111){
    Observable[ 0 ] = m34;
    Observable[ 1 ] = std::fabs(rapidity34) ;
    Observable[ 2 ] = std::fabs(rapidity34) ;
  }
  else if(exptid_.exptid==3112){
    Observable[ 0 ] = m34;
    Observable[ 1 ] = xL ;
  }	
  //Single Top
  else if ((nproc_.nproc>=161)&&(nproc_.nproc<=177)){
  Observable[0]=0.5;}
  // W+top
  else if ((nproc_.nproc>=180)&&(nproc_.nproc<=187))
  { Observable[0]=0.5;}
  //W+charm
  else if(exptid_.exptid==131||exptid_.exptid==132){ Observable[0]=std::fabs(rapidity4);}
  else if(exptid_.exptid==181||exptid_.exptid==182){ Observable[0]=std::fabs(rapidity3);}
  //general
  else if(exptid_.exptid==1581||(nproc_.nproc >= 701 && nproc_.nproc <= 704)) 
  {Observable[0]=pt3; Observable[1]=std::fabs(rapidity3);
//   Observable[2]=pt3; Observable[3]=pt3;  Observable[4]=pt3; Observable[5]=pt3; Observable[6]=pt3;
  }
  else {
    Observable[ 0 ] = std::fabs(rapidity3);
    Observable[ 1 ] = std::fabs(rapidity4);
    Observable[ 2 ] = pt34;
    Observable[ 3 ] = std::fabs(rapidity34);  
    Observable[ 4 ] = m34;
  }
}

int cuts(int igrid)
{
  int fill = 0;
  switch(igrid)
    {
    case(0):
      if ( (nproc_.nproc >= 280) && (nproc_.nproc <= 286) ) { 
	//if ( Observable[1]>100 ) 
         fill = 1;
      }
      else if((nproc_.nproc >= 41) && (nproc_.nproc <= 43)) {
           fill = 1;
      }
      else if((nproc_.nproc >= 11) && (nproc_.nproc <= 19)) {
           fill = 1;
      }
      else fill = 1;
      break;
    case(1):
      if ( (nproc_.nproc >= 280) && (nproc_.nproc <= 286) ) { 
	//if ( std::fabs(Observable[0])<1.37 ) 
           fill = 1;
      }
      else if((nproc_.nproc >= 41) && (nproc_.nproc <= 43)) {
           fill = 1;
      }  
      else fill = 1;
      break;
    case(2):
      if ( (nproc_.nproc >= 280) && (nproc_.nproc <= 286) ) { 
	//if ( std::fabs(Observable[0])>=1.52 && std::fabs(Observable[0])<2.37) 
           fill = 1;
      }
      else if((nproc_.nproc >= 41) && (nproc_.nproc <= 43)) {
        if ( std::fabs(Observable[0])<1.0 )  fill = 1;
      }
/*      else if(exptid_.exptid==1581||(nproc_.nproc >= 701 && nproc_.nproc <= 704)){
        if( std::fabs(Observable[1])>=2.0 && std::fabs(Observable[1])>=2.5) fill=1;
      }
*/
      else fill = 1;
      break;
    case(3):
      if((nproc_.nproc >= 41) && (nproc_.nproc <= 43)) {
        if ( std::fabs(Observable[0])>1.0 && std::fabs(Observable[0])<2.0 ) fill = 1;
      }
/*      else if(exptid_.exptid==1581||(nproc_.nproc >= 701 && nproc_.nproc <= 704)){
        if( std::fabs(Observable[1])>=2.5 && std::fabs(Observable[1])>=3.0) fill=1;
      }
*/
      else fill = 1;
      break;
    case(4):
      if((nproc_.nproc >= 41) && (nproc_.nproc <= 43)) {
        if ( std::fabs(Observable[0])>2.0 && std::fabs(Observable[0])<2.4 )  fill = 1;
      }
/*      else if(exptid_.exptid==1581||(nproc_.nproc >= 701 && nproc_.nproc <= 704)){
        if( std::fabs(Observable[1])>=3.0 && std::fabs(Observable[1])>=3.5) fill=1;
      }
      else fill = 1;
*/
      break;
    case(5):
/*      if(exptid_.exptid==1581||(nproc_.nproc >= 701 && nproc_.nproc <= 704)){
        if( std::fabs(Observable[1])>=3.5 && std::fabs(Observable[1])>=4.0) fill=1;
      }
      else fill = 1;
*/
      break;
    case(6):
/*      if(exptid_.exptid==1581||(nproc_.nproc >= 701 && nproc_.nproc <= 704)){
        if( std::fabs(Observable[1])>=4.0 && std::fabs(Observable[1])>=4.5) fill=1;
      }
      else fill = 1;
*/
      break;
    default: 
      std::cerr<<" In gridwrap.cpp::cuts(int). No such process : "<<igrid<<std::endl;
      std::exit(-1);
    }
  return fill;
}



// namespace mcfm_bridge;

/// function pointer hooks - set to 0 when no functions defined and applgrid not linked
extern void (*book_gridptr)();                         
extern void (*fill_gridptr)(const double evt[][mxpart] );
extern void (*write_gridptr)(double& );   


extern "C" bool setup_mcfmbridge() { 
  std::cout << "setup_mcfmbridge()" << std::endl;
  book_gridptr  = book_grid;
  fill_gridptr  = fill_grid;
  write_gridptr = write_grid;
  return true;
}

extern "C" bool setup_mcfmbridge_() { 
  std::cout << "setup_mcfmbridge()" << std::endl;
  return setup_mcfmbridge();
}
 

bool mcfm_bridge_status = setup_mcfmbridge();


